#ifndef _CONVEX_HULL_4_DIM_H_
#define _CONVEX_HULL_4_DIM_H_

#include <vector>
#include <map>
#include <set>

#include "DArray.h"
#include "Point4.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846264338327950288419716939937510582
#endif

using namespace DartLib;

namespace DartLib
{
namespace ConvexHull4D
{

class CVertex
{
  public:
    CVertex(int idx) : m_idx(idx){};

    int idx() const { return m_idx; };
    CPoint4& point() { return m_point; };

    CVertex& operator=(const CVertex& rhs)
    {
        m_idx = rhs.m_idx;
        m_point = rhs.m_point;
        return *this;
    };

  protected:
    int m_idx;
    CPoint4 m_point;
};

class CFace
{
  public:
    CFace(int vertices[])
    {
        m_vertices[0] = vertices[0];
        m_vertices[1] = vertices[1];
        m_vertices[2] = vertices[2];

        m_hash = compute_hash(vertices);
    };

    size_t hash() const { return m_hash; };

    int vertex(const int i) const { return m_vertices[i]; };

    CFace& operator=(const CFace& rhs)
    {
        for (int i = 0; i < 3; ++i)
            m_vertices[i] = rhs.m_vertices[i];
        m_hash = rhs.m_hash;
        return *this;
    };

  protected:
    size_t compute_hash(int vertices[])
    {
        // bubble sort
        for (int i = 0; i < 2; ++i)
        {
            for (int j = 0; j < 2 - i; ++j)
            {
                if (vertices[j] > vertices[j + 1])
                {
                    int tmp = vertices[j + 1];
                    vertices[j + 1] = vertices[j];
                    vertices[j] = tmp;
                }
            }
        }

        // compute the hash value
        return (size_t) vertices[0] * 1000000000000L + 
               (size_t) vertices[1] * 1000000L + 
               (size_t) vertices[2];
    }

  protected:
    int m_vertices[3];
    size_t m_hash;
};

class CTet
{
  public:
    CTet(int vertices[]) { set_vertices(vertices); };

    void set_vertices(int vertices[])
    {
        for (int i = 0; i < 4; ++i)
            m_vertices[i] = vertices[i];
    };

    int vertex(const int i) const { return m_vertices[i]; };

    CTet& operator=(const CTet& rhs)
    {
        for (int i = 0; i < 4; ++i)
            m_vertices[i] = rhs.m_vertices[i];
        return *this;
    };

  protected:
    int m_vertices[4];
};

class CTetMesh
{
  public:
    CTetMesh(){};

    void load(CDArray<CPoint4>& pts, CDArray<CDArray<int>>& tets);

    int create_vertex(const CPoint4& pt);

    int create_tet(const CDArray<int>& tet);

    void remove_tet(const int tet_idx);

    void write(const char* output);

    CDArray<CVertex>& vertices() { return m_vertices; };
    
    CDArray<CTet>& tets() { return m_tets; };

    std::map<size_t, CFace>& boundary_faces() { return m_boundary_faces; };

  protected:
    CDArray<CVertex> m_vertices;

    CDArray<CTet> m_tets;

    std::map<size_t, CFace> m_boundary_faces;
};

class CConvexHull4D
{
  public:
    /*!
     *  Construct function
     */
    CConvexHull4D();

    /*!
     *  Initialize a group of 4d points with (x, y, z) in a unit ball
     *    randomly, and w = (x^2 + y^2 + z^2) / 2.
     *  \param [in] num_pts: the number of the points.
     */
    void init(int num_pts);

    /*!
     *  Feed sites into CConvexHull
     *  \param [in] sites: a list of 4d points
     */
    void init(std::vector<CPoint4>& sites);

    /*!
     *  Insert one point, the convex hull will be updated if necessary.
     *  \param [in] p: a point which will be inserted.
     */
    void insert(CPoint4& p);

    /*!
     *  Construct the convex hull.
     */
    void construct();

    /*!
     *  The input sites
     *  \return the reference
     */
    CDArray<CPoint4>& sites() { return m_sites; };

    /*!
     *  The convex hull represented by a volume mesh.
     *  \return the reference
     */
    CTetMesh& hull() { return m_pHull; };

    /*!
     *  Determine the sign of the 4d-volume constructed by a tetrahedron 
     *    and a point.
     *  \param [in] pVol: a tetrahedron reference
     *  \param [in]    p: a point reference
     *  \return +1, -1 or 0
     */
    int volume_sign(const CTet& tet, const CPoint4& p);

  protected:
    /*!
     *  Determine whether a point locates in the current convex hull
     *  \param [in] p: a point reference
     *  \return true if p locates in the convex hull, or false
     */
    bool _inside(CPoint4& p);

    /*!
     *  Remove the tetrahedrons which are visiable viewing from point p
     *  \param [in] p: the observer
     */
    void _remove_visiable(CPoint4& p);

    /*!
     *  Close the cap to form a new convex hull
     *  \param [in] p: a point which will connect to the boundary of the cap.
     */
    void _close_cap(CPoint4& p);

  protected:
    /*!
     *  the sites will be used to construct the convex hull
     */
    CDArray<CPoint4> m_sites;

    /*!
     *  A tetrahedral mesh used to hold the convex hull
     */
    CTetMesh m_pHull;
};
} // namespace ConvexHull4D
} // namespace DartLib
#endif //! _CONVEX_HULL_4_DIM_H_
