#ifndef _DARRAY_H_
#define _DARRAY_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

namespace DartLib
{

/*!
 *  A simple dynamic array class.
 *    NOTE: This is designed specially for convex hull algorithm,
 *          please do NOT use it in general case.
 */
template <typename T>
class CDArray
{
public:
    CDArray();
    CDArray(size_t size);
    ~CDArray();

    T operator[](int i) const { return m_data[i]; };

    T& operator[](int i) { return m_data[i]; };

    CDArray<T>& operator=(const CDArray<T>& rhs);

    size_t find(const T& v);

    size_t end() { return m_size; };

    T& back() { return m_data[m_size - 1]; };

    void push(const T& v);

    T& allocate();

    void remove(size_t i);

    void remove(const T& v);

    size_t size() const { return m_size; };
    size_t capacity() const { return m_capacity; };

  protected:
    T* m_data;
    size_t m_size;
    size_t m_capacity;
    size_t m_step_capacity = 4096;
};

template <typename T>
CDArray<T>::CDArray()
{
    m_size = 0;
    m_capacity = 4096;
    m_data = (T*) malloc((size_t)(m_capacity * sizeof(T)));
}

template <typename T>
CDArray<T>::CDArray(size_t size)
{
    m_size = size;
    m_capacity = size;
    m_data = (T*) malloc((size_t)(m_capacity * sizeof(T)));
}

template <typename T>
CDArray<T>::~CDArray()
{
    m_size = 0;
    m_capacity = 0;
    free(m_data);
    m_data = NULL;
}

template <typename T>
CDArray<T>& CDArray<T>::operator=(const CDArray<T>& rhs)
{
    m_size = rhs.m_size;
    m_capacity = rhs.m_capacity;
    m_step_capacity = rhs.m_step_capacity;
    
    m_data = (T*) malloc((size_t)(m_capacity * sizeof(T)));
    memcpy(m_data, rhs.m_data, (size_t)(m_size * sizeof(T)));

    return *this;
}

template <typename T>
size_t CDArray<T>::find(const T& v)
{
    for (size_t i = 0; i < m_size; ++i)
    {
        if (m_data[i] == v)
            return i;
    }

    return m_size;
}

template <typename T>
void CDArray<T>::push(const T& v)
{
    T& data = this->allocate();
    data = v;
}

template <typename T>
T& CDArray<T>::allocate()
{
    if (m_size + 1 <= m_capacity)
    {
        return m_data[m_size++];
    }
    else
    {
        T* data = m_data;
        size_t data_length = (size_t)(m_capacity * sizeof(T));

        m_capacity += m_step_capacity;
        m_data = (T*) malloc((size_t)(m_capacity * sizeof(T)));
        if (m_data == NULL)
        {
            printf("ERROR: allocate memory space failed!\n");
            exit(EXIT_FAILURE);
        }

        // copy
        memcpy(m_data, data, data_length);
        free(data);

        return allocate();
    }
}

template <typename T>
void CDArray<T>::remove(size_t i)
{
    if (i < m_size - 1)
        memcpy(&m_data[i], &m_data[m_size-1], sizeof(T));
    --m_size;
}

template <typename T>
void CDArray<T>::remove(const T& v)
{
    size_t i = find(v);
    if (i != end()) remove(i);
}

} // namespace DartLib
#endif //! _DARRAY_H_
