#ifndef _CONVEX_HULL_4D_MESH_
#define _CONVEX_HULL_4D_MESH_

#include "Geometry/Plane.h"
#include "Mesh/Header.h"
#include "Point4.h"

using namespace DartLib;

namespace DartLib
{

class CMyVertex_3 : public CVertex_3
{
  public:
    CMyVertex_3() : m_outside(false){};

    bool& outside() { return m_outside; };
    
  protected:
    bool m_outside;
};

class CMyFace_3 : public CFace_3
{
  public:
    CMyFace_3() {};

    CPoint& normal() { return m_normal; }
  protected:
    /*!
     *  Face normal, defined based on the darts on the face
     */
    CPoint m_normal;
};

class CMyTet_3 : public CVolume_3
{
  public:
    CMyTet_3() : m_outside(false){};

    bool& outside() { return m_outside; };

  protected:
    bool m_outside;
};

template <typename V, typename E, typename F, typename T, typename Dart>
class TMyTMesh : public TDynamicMesh_3<V, E, F, T, Dart>
{
  public:
    using typename TDynamicMesh_3<V, E, F, T, Dart>::VertexIterator;
    using typename TDynamicMesh_3<V, E, F, T, Dart>::VolumeIterator;
    using typename TDynamicMesh_3<V, E, F, T, Dart>::VolumeVertexIterator;
    using typename TDynamicMesh_3<V, E, F, T, Dart>::FaceIterator;
    using typename TDynamicMesh_3<V, E, F, T, Dart>::FaceVertexIterator;
  
    void normalize();

    CPoint compute_face_normal(F* pF);

    void compute_face_normal();

    void cut(const CPlane& plane);

    std::vector<Dart*>& halffaces_above() { return m_pHalffaces_above; };
    std::vector<Dart*>& halffaces_below() { return m_pHalffaces_below; };
  protected:
    std::vector<Dart*> m_pHalffaces_above;
    std::vector<Dart*> m_pHalffaces_below;
};

using CMyTMesh = TMyTMesh<CMyVertex_3, CEdge_3, CMyFace_3, CMyTet_3, CDart_3>;

template <typename V, typename E, typename F, typename T, typename Dart>
void TMyTMesh<V, E, F, T, Dart>::normalize()
{
    CPoint vmax(-1e+10, -1e+10, -1e+10);
    CPoint vmin(+1e+10, +1e+10, +1e+10);

    for (VertexIterator viter(this); !viter.end(); ++viter)
    {
        V* pV = *viter;
        const CPoint& p = pV->point();

        for (int k = 0; k < 3; k++)
        {
            vmax[k] = (vmax[k] > p[k]) ? vmax[k] : p[k];
            vmin[k] = (vmin[k] < p[k]) ? vmin[k] : p[k];
        }
    }

    CPoint center = (vmax + vmin) / 2.0;

    double d = 0;
    for (VertexIterator viter(this); !viter.end(); ++viter)
    {
        V* pV = *viter;
        CPoint& p = pV->point();
        p = p - center;

        for (int k = 0; k < 3; k++)
        {
            if (fabs(p[k]) > d)
                d = fabs(p[k]);
        }
    }

    for (VertexIterator viter(this); !viter.end(); ++viter)
    {
        V* pV = *viter;
        CPoint& p = pV->point();
        p /= d;
    }
}

template <typename V, typename E, typename F, typename T, typename Dart>
CPoint TMyTMesh<V, E, F, T, Dart>::compute_face_normal(F* pF)
{
    CPoint pts[3];

    int i = 0;
    for (FaceVertexIterator fviter(pF); !fviter.end(); ++fviter)
    {
        V* pV = *fviter;
        pts[i++] = pV->point();
        if (i >= 3)
            break;
    }
    CPoint n = (pts[1] - pts[0]) ^ (pts[2] - pts[0]);
    n /= n.norm();

    return n;
}

template <typename V, typename E, typename F, typename T, typename Dart>
void TMyTMesh<V, E, F, T, Dart>::compute_face_normal()
{
    CPoint pts[3];
    for (FaceIterator fiter(this); !fiter.end(); ++fiter)
    {
        F* pF = *fiter;
        pF->normal() = compute_face_normal(pF);
    }
}

template <typename V, typename E, typename F, typename T, typename Dart>
void TMyTMesh<V, E, F, T, Dart>::cut(const CPlane& plane)
{
    m_pHalffaces_above.clear();
    m_pHalffaces_below.clear();

    for (VertexIterator viter(this); !viter.end(); ++viter)
    {
        V* pV = *viter;
        CPoint& pt = pV->point();
        pV->outside() = (plane.side(pt) >= 0);
    }

    for (VolumeIterator titer(this); !titer.end(); ++titer)
    {
        T* pT = *titer;
        pT->outside() = true;
        for (VolumeVertexIterator tviter(pT); !tviter.end(); ++tviter)
        {
            V* pV = *tviter;
            if (!pV->outside())
            {
                pT->outside() = false;
                break;
            }
        }
    }

    for (FaceIterator fiter(this); !fiter.end(); ++fiter)
    {
        F* pF = *fiter;
        for (int i = 0; i < 2; ++i)
        {
            Dart* pD = (i == 0) ? this->D(pF) : this->beta(3, this->D(pF));
            if (pD == NULL)
                continue;

            T* pT[2] = {this->C3(pD), this->C3(this->beta(3, pD))};

            if (pT[1] == NULL)
            {
                if (pT[0]->outside())
                    m_pHalffaces_above.push_back(pD);
                else
                    m_pHalffaces_below.push_back(pD);

                continue;
            }

            if (pT[0]->outside() && !pT[1]->outside())
            {
                m_pHalffaces_above.push_back(pD);
            }
            else if (!pT[0]->outside() && pT[1]->outside())
            {
                pD = this->beta(3, pD);
                m_pHalffaces_below.push_back(pD);
            }
        }
    }
}

} // namespace DartLib
#endif // !_CONVEX_HULL_4D_MESH_
