#include <time.h>
#include <stdio.h>

#include "ConvexHull4D.h"

namespace DartLib
{
namespace ConvexHull4D
{

void CTetMesh::load(CDArray<CPoint4>& pts, CDArray<CDArray<int>>& tets)
{
    for (size_t i = 0; i < pts.size(); ++i)
    {
        create_vertex(pts[i]);
    }

    for (size_t i = 0; i < tets.size(); ++i)
    {
        create_tet(tets[i]);
    }
}

int CTetMesh::create_vertex(const CPoint4& pt)
{
    int idx = m_vertices.size();
    m_vertices.push(CVertex(idx));
    m_vertices.back().point() = pt;

    return idx;
}

int CTetMesh::create_tet(const CDArray<int>& tet)
{
    // create boundary face
    static int tbl_tet_face[4][3] = {{0, 1, 2}, {0, 2, 3}, {0, 3, 1}, {1, 3, 2}};
    for (int i = 0; i < 4; ++i)
    {
        int face_verts[3];
        for (int j = 0; j < 3; ++j)
        {
            face_verts[j] = tet[tbl_tet_face[i][j]];
        }

        // erase the face if exist, or insert the new face as a boundary face
        CFace f(face_verts);
        auto it = m_boundary_faces.find(f.hash());
        if (it == m_boundary_faces.end())
            m_boundary_faces.insert(std::make_pair(f.hash(), f));
        else
            m_boundary_faces.erase(it);
    }

    // create tet
    int verts[4];
    for (int j = 0; j < 4; ++j)
        verts[j] = tet[j];
    m_tets.push(CTet(verts));

    return m_tets.size() - 1;
}

void CTetMesh::remove_tet(const int tet_idx)
{
    CTet& tet = m_tets[tet_idx];

    // modify the boundary faces
    static int tbl_tet_face[4][3] = {{0, 1, 2}, {0, 2, 3}, {0, 3, 1}, {1, 3, 2}};
    for (int i = 0; i < 4; ++i)
    {
        int face_verts[3];
        for (int j = 0; j < 3; ++j)
        {
            face_verts[j] = tet.vertex(tbl_tet_face[i][j]);
        }

        // if the face does not exist, then insert it as a boundary face
        // invert the orientation of the face
        int tmp = face_verts[0];
        face_verts[0] = face_verts[1], face_verts[1] = tmp;

        CFace f(face_verts);
        auto it = m_boundary_faces.find(f.hash());
        if (it == m_boundary_faces.end())
            m_boundary_faces.insert(std::make_pair(f.hash(), f));
        else
            m_boundary_faces.erase(it);
    }

    // remove tet
    m_tets.remove((size_t) tet_idx);
}

void CTetMesh::write(const char* output)
{
    FILE* fp;
    fp = fopen(output, "w");
    if (!fp)
    {
        printf("open file %s failed\n", output);
        exit(0);
    }

    for (int i = 0; i < m_vertices.size(); ++i)
    {
        auto& v = m_vertices[i];
        fprintf(fp, "Vertex %d %g %g %g\n", v.idx() + 1, v.point()[0], v.point()[1], v.point()[2]);
    }

    for (int i = 0; i < m_tets.size(); ++i)
    {
        auto& tet = m_tets[i];

        fprintf(fp, "Tet %d", i + 1);
        for (int j = 0; j < 4; ++j)
            fprintf(fp, " %d", tet.vertex(j) + 1);
        fprintf(fp, "\n");
    }

    fclose(fp);
}

CConvexHull4D::CConvexHull4D() {}

void CConvexHull4D::init(int num_pts)
{
    std::vector<CPoint4> sites;

    // 1. randomly generate the input points
    srand((unsigned) time(NULL));
    double n[3];
    for (int i = 0; i < num_pts; ++i)
    {
        n[0] = rand() / double(RAND_MAX);
        n[1] = rand() / double(RAND_MAX);
        n[2] = 2.0 * M_PI * rand() / double(RAND_MAX);

        // refer: http://datagenetics.com/blog/january32020/index.html
        double r = std::pow(n[0], 1.0 / 3.0);
        double phi = std::acos((2.0 * n[1]) - 1.0);
        double theta = n[2];
        double x = r * std::sin(phi) * std::cos(theta);
        double y = r * std::sin(phi) * std::sin(theta);
        double z = r * std::cos(phi);
        double w = (x * x + y * y + z * z) / 2.0;

        CPoint4 p(x, y, z, w);
        sites.push_back(p);
    }

    // 2. initialize the input sites
    init(sites);
}

void CConvexHull4D::init(std::vector<CPoint4>& sites)
{
    for (auto& site : sites)
        m_sites.push(site);

    // construst an initial convex hull with double tetrahedrons
    for (int i = 0; i < 4; ++i)
    {
        CPoint4& p = m_sites[i];
        m_pHull.create_vertex(p);
    }

    CDArray<int> tet_verts(4);
    tet_verts[0] = 0, tet_verts[1] = 1, tet_verts[2] = 2, tet_verts[3] = 3;
    m_pHull.create_tet(tet_verts);
    tet_verts[0] = 0, tet_verts[1] = 2, tet_verts[2] = 1, tet_verts[3] = 3;
    m_pHull.create_tet(tet_verts);
}

void CConvexHull4D::insert(CPoint4& p)
{
    if (!_inside(p))
    {
        _remove_visiable(p);
        _close_cap(p);
    }
}

void CConvexHull4D::construct()
{
    printf("Constructing...\n");
    for (size_t i = 4; i < m_sites.size(); ++i)
    {
        auto& p = m_sites[i];
        insert(p);
        printf("\r%.2f%%", 100.0 * (i + 1) / m_sites.size());
    }
    printf("\n");
    printf("Done.\n");
}

int CConvexHull4D::volume_sign(const CTet& tet, const CPoint4& p)
{
    double vol;
    double ax, ay, az, aw, bx, by, bz, bw, cx, cy, cz, cw, dx, dy, dz, dw;

    int v[4];
    for (int i = 0; i < 4; ++i)
        v[i] = tet.vertex(i);

    // I: the I-st vertex of the tet.
    // J: the J-st coordinate component.
    const auto& vertices = m_pHull.vertices();
#define DIFF(I, J) vertices[v[I]].point()[J] - p[J]

    ax = DIFF(0, 0);
    ay = DIFF(0, 1);
    az = DIFF(0, 2);
    aw = DIFF(0, 3);

    bx = DIFF(1, 0);
    by = DIFF(1, 1);
    bz = DIFF(1, 2);
    bw = DIFF(1, 3);

    cx = DIFF(2, 0);
    cy = DIFF(2, 1);
    cz = DIFF(2, 2);
    cw = DIFF(2, 3);

    dx = DIFF(3, 0);
    dy = DIFF(3, 1);
    dz = DIFF(3, 2);
    dw = DIFF(3, 3);

#undef DIFF

    // This version of computing determinant is faster then
    //   the following version. It contains 23 addition operations
    //   and 40 multiplication operations.
    vol = ((dx * cy - cx * dy) * bz +
           (bx * dy - dx * by) * cz + 
           (cx * by - bx * cy) * dz
          ) * aw +

          ((cx * dy - dx * cy) * az + 
           (dx * ay - ax * dy) * cz + 
           (ax * cy - cx * ay) * dz
          ) * bw +

          ((dx * by - bx * dy) * az + 
           (ax * dy - dx * ay) * bz + 
           (bx * ay - ax * by) * dz
          ) * cw +

          ((bx * cy - cx * by) * az + 
           (cx * ay - ax * cy) * bz + 
           (ax * by - bx * ay) * cz
          ) * dw;

    // This version contains 23 addition operations
    //   and 72 multiplication operations.
    /*
    vol = dx * cy * bz * aw - cx * dy * bz * aw -
          dx * by * cz * aw + bx * dy * cz * aw +
          cx * by * dz * aw - bx * cy * dz * aw -
          dx * cy * az * bw + cx * dy * az * bw +
          dx * ay * cz * bw - ax * dy * cz * bw -
          cx * ay * dz * bw + ax * cy * dz * bw +
          dx * by * az * cw - bx * dy * az * cw -
          dx * ay * bz * cw + ax * dy * bz * cw +
          bx * ay * dz * cw - ax * by * dz * cw -
          cx * by * az * dw + bx * cy * az * dw +
          cx * ay * bz * dw - ax * cy * bz * dw -
          bx * ay * cz * dw + ax * by * cz * dw;
    */

    if (vol > +0.0)
        return +1;
    else if (vol < -0.0)
        return -1;
    else
        return 0;
}

bool CConvexHull4D::_inside(CPoint4& p)
{
    for (size_t i = 0; i < m_pHull.tets().size(); ++i)
    {
        const auto& tet = m_pHull.tets()[i];
        if (volume_sign(tet, p) <= 0) // outside
            return false;
    }
    return true;
}

void CConvexHull4D::_remove_visiable(CPoint4& p)
{
    int sign = 0;
    // We use decrement here to avoid the `dirty` index.
    for (int i = m_pHull.tets().size() - 1; i >= 0; --i)
    {
        const auto& tet = m_pHull.tets()[i];
        sign = volume_sign(tet, p);
        if (sign <= 0)
            m_pHull.remove_tet(i);
    }
}

void CConvexHull4D::_close_cap(CPoint4& p)
{
    // insert new vertex
    int vidx = m_pHull.create_vertex(p);

    // collect new tetrahedrons
    CDArray<CDArray<int>> tets;
    for (auto& face : m_pHull.boundary_faces())
    {
        CDArray<int> verts(4);
        for (int i = 0; i < 3; ++i)
            verts[i] = face.second.vertex(i);
        verts[3] = vidx;

        // invert the orientation of the tetrahedron
        int tmp = verts[0];
        verts[0] = verts[1], verts[1] = tmp;

        // collected new tetrahedrons
        tets.push(verts);
    }

    // insert new tetrahedrons
    for (int i = 0; i < tets.size(); ++i)
    {
        const auto& verts = tets[i];
        m_pHull.create_tet(verts);
    }
}

} // namespace ConvexHull4D
} // namespace DartLib
