#include <iostream>
#include <unordered_set>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef MAC_OS
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif // MAC_OS

#include "Viewer/Arcball.h" /* Arc Ball Interface */

#include "MyTMesh.h"
#include "ConvexHull4D.h"

using namespace DartLib;
using namespace DartLib::ConvexHull4D;

/* window width and height */
int win_width, win_height;
int gButton;
int startx, starty;
int shadeFlag = 0;
int displayMode = 0;
int displayModel = 1; // mesh or tmesh
bool switchDirection = true;

/* rotation quaternion and translation vector for the object */
CQrot ObjRot(0, 0, 1, 0);
CPoint ObjTrans(0, 0, 0);

/* global convexhull object */
CConvexHull4D g_ch;

/* tet mesh projected from the 4d convex hull */
CMyTMesh g_tmesh;

/* cut plane */
double d = 0;
CPoint n(0, 0, 1);

/* arcball object */
CArcball arcball;

void build_tmesh() 
{
    std::vector<CPoint> vertices;
    std::vector<std::vector<int>> volumes;

    auto& verts = g_ch.hull().vertices();
    auto& tets = g_ch.hull().tets();

    for (int i = 0; i < verts.size(); ++i)
    {
        auto& p = verts[i].point();
        vertices.push_back(CPoint(p[0], p[1], p[2]));
    }

    const CPoint4 w_max(0, 0, 0, +1e100);
    for (int i = 0; i < tets.size(); ++i)
    {
        const auto& tet = g_ch.hull().tets()[i];
        if (g_ch.volume_sign(tet, w_max) < 0)
            continue;

        std::vector<int> tet_verts(4);
        for (int j = 0; j < 4; ++j)
            tet_verts[j] = tets[i].vertex(j);

        volumes.push_back(tet_verts);
    }

    g_tmesh.load(vertices, volumes);
    g_tmesh.compute_face_normal();
}

/*! setup the object, transform from the world to the object coordinate system */
void setupObject(void)
{
    double rot[16];

    glTranslated(ObjTrans[0], ObjTrans[1], ObjTrans[2]);
    ObjRot.convert(rot);
    glMultMatrixd((GLdouble*) rot);
}

/*! the eye is always fixed at world z = +5 */

void setupEye(void)
{
    glLoadIdentity();
    gluLookAt(0, 0, 5, 0, 0, 0, 0, 1, 0);
}

/*! setup light */
void setupLight()
{
    CPoint position(0, 0, 1);
    GLfloat lightOnePosition[4] = {(float)position[0], (float)position[1], (float)position[2], 0.f};
    glLightfv(GL_LIGHT1, GL_POSITION, lightOnePosition);
}

void draw_sites() 
{
    glDisable(GL_LIGHTING);
    glPointSize(2.0);
    glBegin(GL_POINTS);
    glColor3d(0.85, 0.85, 0.85);

    for (int i = 0; i < g_ch.sites().size(); ++i)
    {
        CPoint4& p = g_ch.sites()[i];
        glVertex3d(p[0], p[1], p[2]);
    }

    glEnd();
}

void draw_face(CMyTMesh::CFace* face, bool inv = false)
{
    CPoint n = face->normal();

    std::vector<CMyTMesh::CVertex*> verts;
    for (CMyTMesh::FaceVertexIterator fviter(face); !fviter.end(); ++fviter)
    {
        CMyTMesh::CVertex* pV = *fviter;
        verts.push_back(pV);
    }

    if (inv)
    {
        std::reverse(verts.begin(), verts.end());
        n = n * (-1);
    }

    glBegin(GL_TRIANGLES);
    for (auto& pV : verts)
    {
        CPoint& p = pV->point();
        if (!switchDirection)
            glNormal3d(n[0], n[1], n[2]);
        else
            glNormal3d(-n[0], -n[1], -n[2]);
        glVertex3d(p[0], p[1], p[2]);
    }
    glEnd();
}

void draw_halffaces(std::vector<CMyTMesh::CDart*>& darts, GLenum mode)
{
    glEnable(GL_LIGHTING);
    glColor3f(1.0, 0.5, 0.0);
    glPolygonMode(GL_FRONT, mode);

    for (CMyTMesh::CDart* pD : darts)
    {
        auto f = g_tmesh.dart_face(pD);
        CMyTMesh::CDart* darts_on_f[3];
        darts_on_f[0] = g_tmesh.face_dart(f);
        darts_on_f[1] = g_tmesh.beta(1, darts_on_f[0]);
        darts_on_f[2] = g_tmesh.beta(1, darts_on_f[1]);

        if (darts_on_f[0] == pD || darts_on_f[1] == pD || darts_on_f[2] == pD)
        {
            // TODO: I think `draw_face(f, false)` should work well here,
            //       but I must do as follows.
            if (!g_tmesh.boundary(pD)) // inner face
                draw_face(f, true);
            else
                draw_face(f, false);
        }
        else
        {
            // TODO: I think `draw_face(f, true)` should work well here,
            //       but I must do as follows.
            if (!g_tmesh.boundary(pD)) // inner face
                draw_face(f, false);
            else
                draw_face(f, true);
        }
    }
}

void draw_plane() 
{
    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glPolygonMode(GL_FRONT, GL_FILL);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glColor4d(0.8, 0.8, 0.8, 0.3);

    CPoint corners[4];
    if (n[0] == 1)
    {
        corners[0] = CPoint(d, +1.2, -1.2);
        corners[1] = CPoint(d, +1.2, +1.2);
        corners[2] = CPoint(d, -1.2, +1.2);
        corners[3] = CPoint(d, -1.2, -1.2);
    }
    else if (n[1] == 1)
    {
        corners[0] = CPoint(-1.2, d, +1.2);
        corners[1] = CPoint(+1.2, d, +1.2);
        corners[2] = CPoint(+1.2, d, -1.2);
        corners[3] = CPoint(-1.2, d, -1.2);
    }
    else if (n[2] == 1)
    {
        corners[0] = CPoint(+1.2, -1.2, d);
        corners[1] = CPoint(+1.2, +1.2, d);
        corners[2] = CPoint(-1.2, +1.2, d);
        corners[3] = CPoint(-1.2, -1.2, d);
    }

    glBegin(GL_POLYGON);
    glNormal3d(n[0], n[1], n[2]);
    for (int i = 0; i < 4; ++i)
    {
        glVertex3d(corners[i][0], corners[i][1], corners[i][2]);
    }
    glEnd();

    glEnable(GL_CULL_FACE);
    glDisable(GL_BLEND);
}

/*! draw mesh */
void draw_mesh()
{
    draw_sites();

    switch (displayMode)
    {
        case 0:
            draw_halffaces(g_tmesh.halffaces_below(), GL_FILL);
            break;
        case 1:
            draw_halffaces(g_tmesh.halffaces_below(), GL_LINE);
            break;
        case 2:
            draw_halffaces(g_tmesh.halffaces_below(), GL_FILL);
            draw_halffaces(g_tmesh.halffaces_above(), GL_LINE);
            break;
    }

    draw_plane();
}

/*! display call back function
 */
void display()
{
    /* clear frame buffer */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    setupLight();
    /* transform from the eye coordinate system to the world system */
    setupEye();
    glPushMatrix();
    /* transform from the world to the ojbect coordinate system */
    setupObject();
    /* draw the mesh */
    draw_mesh();

    glPopMatrix();
    glutSwapBuffers();
}

/*! Called when a "resize" event is received by the window. */

void reshape(int w, int h)
{
    float ar;

    win_width = w;
    win_height = h;

    ar = (float) (w) / h;
    glViewport(0, 0, w, h); /* Set Viewport */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    // magic imageing commands
    gluPerspective(40.0, /* field of view in degrees */
                   ar,   /* aspect ratio */
                   1.0,  /* Z near */
                   100.0 /* Z far */);

    glMatrixMode(GL_MODELVIEW);

    glutPostRedisplay();
}

/*! helper function to remind the user about commands, hot keys */
void help()
{
    printf("i  -  insert site one by one\n");
    printf("I  -  insert the remaining sites\n");
    printf("C  -  construct the convex hull in one step\n");
    printf("g  -  switch display mode\n");
    printf("+  -  shfit the cutting plane upwards\n");
    printf("-  -  shfit the cutting plane downwards\n");
    printf("x  -  set the cutting plane normal (1,0,0)\n");
    printf("y  -  set the cutting plane normal (0,1,0)\n");
    printf("z  -  set the cutting plane normal (0,0,1)\n");
    printf("?  -  help Information\n");
    printf("esc - quit\n");
}

/*! Keyboard call back function */

void keyBoard(unsigned char key, int x, int y)
{
    static int site_index = 4;
    CPoint4 site;
    auto& sites = g_ch.sites();
    site_index = site_index % sites.size();
    site = sites[site_index];

    clock_t begin = clock(), end;
    switch (key)
    {
        case 'i':
            // take next one step
            printf("inserting site: %d (%g %g %g %g)\n", site_index + 1, site[0], site[1], site[2], site[3]);
            g_ch.insert(site);
            ++site_index;

            build_tmesh();
            g_tmesh.cut(CPlane(n, d));

            break;
        case 'I':
            // take the remaining steps
            for (site_index; site_index < sites.size(); ++site_index)
            {
                site = sites[site_index];
                printf("inserting site: %d (%g %g %g %g)\n", site_index + 1, site[0], site[1], site[2], site[3]);
                g_ch.insert(site);
            }
            end = clock();
            printf("Total time: %g sec.\n", (double)(end - begin) / CLOCKS_PER_SEC);

            build_tmesh();
            g_tmesh.cut(CPlane(n, d));

            break;
        case 'C':
            // construct the convex hull in one step
            g_ch.construct();
            end = clock();
            printf("Total time: %g sec.\n", (double) (end - begin) / CLOCKS_PER_SEC);

            build_tmesh();
            g_tmesh.cut(CPlane(n, d));

            break;
        case '+':
            d += 0.05;
            g_tmesh.cut(CPlane(n, d));
            break;
        case '-':
            d -= 0.05;
            g_tmesh.cut(CPlane(n, d));
            break;
        case 'x':
            n = CPoint(1, 0, 0);
            g_tmesh.cut(CPlane(n, d));
            break;
        case 'y':
            n = CPoint(0, 1, 0);
            g_tmesh.cut(CPlane(n, d));
            break;
        case 'z':
            n = CPoint(0, 0, 1);
            g_tmesh.cut(CPlane(n, d));
            break;
        case 'g':
            displayMode = (displayMode + 1) % 3;
            break;
        case 'r':
            switchDirection = !switchDirection;
            if (switchDirection)
                glFrontFace(GL_CW);
            else
                glFrontFace(GL_CCW);
            break;
        case '?':
            help();
            break;
        case 27:
            exit(0);
            break;
    }

    glutPostRedisplay();
}

/*! setup GL states */
void setupGLstate()
{
    GLfloat lightOneColor[] = {1.f, 1.f, 1.f, 1.f};
    GLfloat globalAmb[] = {.1f, .1f, .1f, 1.f};
    GLfloat lightOnePosition[] = {.0f, .0f, 1.f, 0.f};

    glEnable(GL_CULL_FACE);
    if (switchDirection)
        glFrontFace(GL_CW);
    else
        glFrontFace(GL_CCW);

    glEnable(GL_DEPTH_TEST);
    glClearColor(0.3f, 0.3f, 0.8f, 0.3f);
    glShadeModel(GL_SMOOTH);

    glEnable(GL_LIGHT1);
    glEnable(GL_LIGHTING);
    glEnable(GL_NORMALIZE);
    glEnable(GL_COLOR_MATERIAL);

    glLightfv(GL_LIGHT1, GL_DIFFUSE, lightOneColor);
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, globalAmb);

    glLightfv(GL_LIGHT1, GL_POSITION, lightOnePosition);

    GLfloat diffuseMaterial[4] = {0.5, 0.5, 0.5, 1.0};
    GLfloat mat_specular[] = {1.0, 1.0, 1.0, 1.0};

    glEnable(GL_DEPTH_TEST);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuseMaterial);
    glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
    glMaterialf(GL_FRONT, GL_SHININESS, 25.0);
    glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
}

/*! mouse click call back function */
void mouseClick(int button, int state, int x, int y)
{
    /* set up an arcball around the Eye's center
            switch y coordinates to right handed system  */

    if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
    {
        gButton = GLUT_LEFT_BUTTON;
        arcball = CArcball(win_width, win_height, x - win_width / 2, win_height - y - win_height / 2);
    }

    if (button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN)
    {
        startx = x;
        starty = y;
        gButton = GLUT_MIDDLE_BUTTON;
    }

    if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN)
    {
        startx = x;
        starty = y;
        gButton = GLUT_RIGHT_BUTTON;
    }
    return;
}

/*! mouse motion call back function */
void mouseMove(int x, int y)
{
    CPoint trans;
    CQrot rot;

    /* rotation, call arcball */
    if (gButton == GLUT_LEFT_BUTTON)
    {
        rot = arcball.update(x - win_width / 2, win_height - y - win_height / 2);
        ObjRot = rot * ObjRot;
        glutPostRedisplay();
    }

    /*xy translation */
    if (gButton == GLUT_MIDDLE_BUTTON)
    {
        double scale = 10. / win_height;
        trans = CPoint(scale * (x - startx), scale * (starty - y), 0);
        startx = x;
        starty = y;
        ObjTrans = ObjTrans + trans;
        glutPostRedisplay();
    }

    /* zoom in and out */
    if (gButton == GLUT_RIGHT_BUTTON)
    {
        double scale = 10. / win_height;
        trans = CPoint(0, 0, scale * (starty - y));
        startx = x;
        starty = y;
        ObjTrans = ObjTrans + trans;
        glutPostRedisplay();
    }
}

int main(int argc, char* argv[])
{
    int nSites = 1000;
    if (argc > 1)
        nSites = atoi(argv[1]);
    
    g_ch.init(nSites);
    build_tmesh();
    g_tmesh.cut(CPlane(n, d));

    /* glut stuff */
    glutInit(&argc, argv); /* Initialize GLUT */
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
    glutInitWindowSize(600, 600);
    glutCreateWindow("Tet Mesh Viewer"); /* Create window with given title */
    glViewport(0, 0, 800, 800);

    glutDisplayFunc(display); /* Set-up callback functions */
    glutReshapeFunc(reshape);
    glutMouseFunc(mouseClick);
    glutMotionFunc(mouseMove);
    glutKeyboardFunc(keyBoard);
    setupGLstate();

    glutMainLoop(); /* Start GLUT event-processing loop */

    return 0;
}
