#ifndef _POWER_DIAGRAM_3D_H_
#define _POWER_DIAGRAM_3D_H_

#include "DelaunayTMesh.h"

using namespace DartLib;

namespace DartLib
{
namespace PowerDiagram3D
{

/*!
 *  \class CPowerDiagram3D PowerDiagram3D.h "PowerDiagram3D.h"
 *  \brief CPowerDiagram3D, is used to compute power diagram of 3d points.
 *         1. It lifts the plannar points onto paraboloid w = (x^2 + y^2 + z^2) / 2.
 *         2. A 4d convex hull algorithm is applied on them.
 *         3. The upward faces are removed, and the remaining part projected
 *            onto the 3d Euclidean space constructs the delaunay triangulation 
 *            of the points.
 *         4. The dual of the delaunay triangulation is the voronoi diagram.
 */
class CPowerDiagram3D
{
  public:
    /*!
     *  Initial some random points on unit disk
     *  \param [in] num_pts: number of points
     */
    void init(int num_pts);

    /*!
     *  Feed points into CPowerDiagram3D
     *  \param [in] sites: a list of pointers of points
     */
    void init(const std::vector<CPoint>& points);

    /*!
     *  Compute delaunay triangulation
     */
    void calc_delaunay();

    /*!
     *  Compute voronoi diagram - the dual of the delaunay triangulation
     */
    void calc_voronoi();

    /*!
     *  Reference of the input sites
     *  \return the reference of the array of the input sites
     */
    std::vector<CPoint>& sites() { return m_pts; };

    /*!
     *  Reference of the mesh used to store delaunay triangulation and its dual.
     *  \return the reference of the mesh
     */
    CDelaunayTMesh& mesh() { return m_mesh; };

  protected:
    void _tet_circumsphere(CDelaunayTMesh::CVolume* tet, CPoint& center, double& radius);

    /*!
     *  Computing the determinant for 4x4 matrix
     *  \param [in] mat: a pointer of 4x4 matrix with row major order
     *  \return the determinant
     */
    double _det(const double* mat);

  protected:
    /*!
     *  The input points
     */
    std::vector<CPoint> m_pts;

    /*!
     *  Used to store delauany trianle mesh and the dual mesh - voronoi diagram
     */
    CDelaunayTMesh m_mesh;
};

} // PowerDiagram3D

} // namespace DartLib
#endif // !_POWER_DIAGRAM_3D_H_
