#include "PowerDiagram3D.h"
#include "ConvexHull4D.h"

namespace DartLib
{
namespace PowerDiagram3D
{
void CPowerDiagram3D::init(int num_pts) 
{
    std::vector<CPoint> sites;

    // 1. randomly generate the input points
    srand((unsigned) time(NULL));
    double n[3];
    for (int i = 0; i < num_pts; ++i)
    {
        n[0] = rand() / double(RAND_MAX);
        n[1] = rand() / double(RAND_MAX);
        n[2] = 2.0 * M_PI * rand() / double(RAND_MAX);

        // refer: http://datagenetics.com/blog/january32020/index.html
        double r = std::pow(n[0], 1.0 / 3.0);
        double phi = std::acos((2.0 * n[1]) - 1.0);
        double theta = n[2];
        double x = r * std::sin(phi) * std::cos(theta);
        double y = r * std::sin(phi) * std::sin(theta);
        double z = r * std::cos(phi);

        CPoint p(x, y, z);
        sites.push_back(p);
    }

    // 2. initialize the input sites
    init(sites);
}

void CPowerDiagram3D::init(const std::vector<CPoint>& points) 
{
    m_pts.clear();
    for (const auto p : points)
    {
        m_pts.push_back(p);
    }
}

void CPowerDiagram3D::calc_delaunay() 
{
    // 0. lift the sites
    std::vector<CPoint4> sites;
    for (const auto& p : m_pts)
    {
        double w = 0.5 * (p[0] * p[0] + p[1] * p[1] + p[2] * p[2]);
        CPoint4 s(p[0], p[1], p[2], w);
        sites.push_back(s);
    }

    // 1. compute the convex hull
    ConvexHull4D::CConvexHull4D ch;
    ch.init(sites);
    ch.construct();

    // 2. build the delaunay tetrahedral mesh
    std::vector<CPoint> vertices;
    std::vector<std::vector<int>> volumes;

    auto& verts = ch.hull().vertices();
    auto& tets  = ch.hull().tets();

    for (int i = 0; i < verts.size(); ++i)
    {
        auto& p = verts[i].point();
        vertices.push_back(CPoint(p[0], p[1], p[2]));
    }

    const CPoint4 w_max(0, 0, 0, +1e100);
    for (int i = 0; i < tets.size(); ++i)
    {
        const auto& tet = ch.hull().tets()[i];
        if (ch.volume_sign(tet, w_max) < 0)
            continue;

        std::vector<int> tet_verts(4);
        for (int j = 0; j < 4; ++j)
            tet_verts[j] = tets[i].vertex(j);

        volumes.push_back(tet_verts);
    }

    m_mesh.load(vertices, volumes);
    m_mesh.compute_face_normal(); // for visualization
}

void CPowerDiagram3D::calc_voronoi() 
{
    using M = CDelaunayTMesh;
    // compute the voronoi diagram, the dual of the delaunay tetrahedral mesh
    for (M::VolumeIterator vol_it(&m_mesh); !vol_it.end(); ++vol_it)
    {
        M::CVolume* pTet = *vol_it;
        _tet_circumsphere(pTet, pTet->dual(), pTet->radius());
    }
}

void CPowerDiagram3D::_tet_circumsphere(CDelaunayTMesh::CVolume* pT, 
                                        CPoint& center, double& radius)
{
    using M = CDelaunayTMesh;
    M::CVertex* v[4];
    int i = 0;
    for (M::VolumeVertexIterator vviter(pT); !vviter.end(); ++vviter)
    {
        v[i++] = *vviter;
    }

    const CPoint& p1 = v[0]->point();
    const CPoint& p2 = v[1]->point();
    const CPoint& p3 = v[2]->point();
    const CPoint& p4 = v[3]->point();

    //insert your code here, compute the circum sphere of four points
    //assign the sphere center to center
    //and radius to radius
}

double CPowerDiagram3D::_det(const double* mat)
{
    double m00 = mat[0*4+0], m01 = mat[0*4+1], m02 = mat[0*4+2], m03 = mat[0*4+3];
    double m10 = mat[1*4+0], m11 = mat[1*4+1], m12 = mat[1*4+2], m13 = mat[1*4+3];
    double m20 = mat[2*4+0], m21 = mat[2*4+1], m22 = mat[2*4+2], m23 = mat[2*4+3];
    double m30 = mat[3*4+0], m31 = mat[3*4+1], m32 = mat[3*4+2], m33 = mat[3*4+3];
    
    double det =
       ((m30 * m21 - m20 * m31) * m12 +
        (m10 * m31 - m30 * m11) * m22 + 
        (m20 * m11 - m10 * m21) * m32
       ) * m03 +

       ((m20 * m31 - m30 * m21) * m02 + 
        (m30 * m01 - m00 * m31) * m22 + 
        (m00 * m21 - m20 * m01) * m32
       ) * m13 +

       ((m30 * m11 - m10 * m31) * m02 + 
        (m00 * m31 - m30 * m01) * m12 + 
        (m10 * m01 - m00 * m11) * m32
       ) * m23 +

       ((m10 * m21 - m20 * m11) * m02 + 
        (m20 * m01 - m00 * m21) * m12 + 
        (m00 * m11 - m10 * m01) * m22
       ) * m33;

    return det; 
}

}
}


