#include <iostream>
#include <unordered_set>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef MAC_OS
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif // MAC_OS

#include "Viewer/Arcball.h" /* Arc Ball Interface */

#include "PowerDiagram3D.h"

using namespace DartLib;
using namespace DartLib::PowerDiagram3D;

/* window width and height */
int win_width, win_height;
int gButton;
int startx, starty;
int shadeFlag = 0;
int displayMode = 1;
int displayModel = 1; // mesh or tmesh
bool switchDirection = true;

/* rotation quaternion and translation vector for the object */
CQrot ObjRot(0, 0, 1, 0);
CPoint ObjTrans(0, 0, 0);

/* global power diagram object */
CPowerDiagram3D g_pd;

/* delaunay triangulation */
PowerDiagram3D::CDelaunayTMesh & g_tmesh = g_pd.mesh();

/* the selected dual cell id to display */
int g_selected_vertex_id = 1;
bool g_show_sites = true;
bool g_show_voronoi_vertices = false;

/* cut plane */
double d = 0;
CPoint n(0, 0, 1);

/* arcball object */
CArcball arcball;

/*! setup the object, transform from the world to the object coordinate system */
void setupObject(void)
{
    double rot[16];

    glTranslated(ObjTrans[0], ObjTrans[1], ObjTrans[2]);
    ObjRot.convert(rot);
    glMultMatrixd((GLdouble*) rot);
}

/*! the eye is always fixed at world z = +5 */

void setupEye(void)
{
    glLoadIdentity();
    gluLookAt(0, 0, 5, 0, 0, 0, 0, 1, 0);
}

/*! setup light */
void setupLight()
{
    CPoint position(0, 0, 1);
    GLfloat lightOnePosition[4] = {(float)position[0], (float)position[1], (float)position[2], 0.f};
    glLightfv(GL_LIGHT1, GL_POSITION, lightOnePosition);
}

void draw_sites() 
{
    glDisable(GL_LIGHTING);
    glPointSize(1.0);
    glBegin(GL_POINTS);
    glColor3d(0.85, 0.85, 0.85);

    for (int i = 0; i < g_pd.sites().size(); ++i)
    {
        CPoint& p = g_pd.sites()[i];
        glVertex3d(p[0], p[1], p[2]);
    }

    glEnd();
}

void draw_voronoi_vertices() 
{
    glDisable(GL_LIGHTING);
    glPointSize(1.0);
    glBegin(GL_POINTS);
    glColor3d(0.85, 0.85, 0.0);

    for (CDelaunayTMesh::VolumeIterator vol_it(&g_tmesh); !vol_it.end(); ++vol_it)
    {
        auto pVol = *vol_it;
        CPoint& dual = pVol->dual();
        glVertex3d(dual[0], dual[1], dual[2]);
    }

    glEnd();
}

void draw_voronoi_cells() 
{
    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    //glPolygonMode(GL_FRONT, GL_FILL);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glColor4d(1.0, 1.0, 0.0, 1.0);

    if (g_selected_vertex_id < 0)
    {
        for (CDelaunayTMesh::VertexIterator vit(&g_tmesh); !vit.end(); ++vit)
        {
            auto pV = *vit;

            auto dual_cell = g_tmesh.vertex_dual_cell(pV);
            for (const auto& dual_face : dual_cell)
            {
                CPoint n = g_tmesh.dual_face_normal(dual_face);
                glBegin(GL_POLYGON);
                glNormal3d(n[0], n[1], n[2]);
                for (const auto& p : dual_face)
                {
                    glVertex3d(p[0], p[1], p[2]);
                }
                glEnd();
            }
        }
    }
    else
    { 
        for (CDelaunayTMesh::VertexIterator vit(&g_tmesh); !vit.end(); ++vit)
        {
            auto pV = *vit;
            if (pV->id() != g_selected_vertex_id) continue;

            auto dual_cell = g_tmesh.vertex_dual_cell(pV);
            for (const auto& dual_face : dual_cell)
            {
                CPoint n = g_tmesh.dual_face_normal(dual_face);
                glBegin(GL_POLYGON);
                glNormal3d(n[0], n[1], n[2]);
                for (const auto& p : dual_face)
                {
                    glVertex3d(p[0], p[1], p[2]);
                }
                glEnd();
            }
        }
    }

    glEnable(GL_CULL_FACE);
    glDisable(GL_BLEND);
}

void draw_face(CDelaunayTMesh::CFace* face, bool inv = false)
{
    CPoint n = face->normal();

    std::vector<CDelaunayTMesh::CVertex*> verts;
    for (CDelaunayTMesh::FaceVertexIterator fviter(face); !fviter.end(); ++fviter)
    {
        CDelaunayTMesh::CVertex* pV = *fviter;
        verts.push_back(pV);
    }

    if (inv)
    {
        std::reverse(verts.begin(), verts.end());
        n = n * (-1);
    }

    glBegin(GL_TRIANGLES);
    for (auto& pV : verts)
    {
        CPoint& p = pV->point();
        if (!switchDirection)
            glNormal3d(n[0], n[1], n[2]);
        else
            glNormal3d(-n[0], -n[1], -n[2]);
        glVertex3d(p[0], p[1], p[2]);
    }
    glEnd();
}

void draw_halffaces(std::vector<CDelaunayTMesh::CDart*>& darts, GLenum mode)
{
    glEnable(GL_LIGHTING);
    glColor3f(1.0, 0.5, 0.0);
    glPolygonMode(GL_FRONT, mode);

    for (CDelaunayTMesh::CDart* pD : darts)
    {
        auto f = g_tmesh.dart_face(pD);
        CDelaunayTMesh::CDart* darts_on_f[3];
        darts_on_f[0] = g_tmesh.face_dart(f);
        darts_on_f[1] = g_tmesh.beta(1, darts_on_f[0]);
        darts_on_f[2] = g_tmesh.beta(1, darts_on_f[1]);

        if (darts_on_f[0] == pD || darts_on_f[1] == pD || darts_on_f[2] == pD)
        {
            // TODO: I think `draw_face(f, false)` should work well here,
            //       but I must do as follows.
            if (!g_tmesh.boundary(pD)) // inner face
                draw_face(f, true);
            else
                draw_face(f, false);
        }
        else
        {
            // TODO: I think `draw_face(f, true)` should work well here,
            //       but I must do as follows.
            if (!g_tmesh.boundary(pD)) // inner face
                draw_face(f, false);
            else
                draw_face(f, true);
        }
    }
}

void draw_plane() 
{
    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glColor4d(0.8, 0.8, 0.8, 0.3);

    CPoint corners[4];
    if (n[0] == 1)
    {
        corners[0] = CPoint(d, +1.2, -1.2);
        corners[1] = CPoint(d, +1.2, +1.2);
        corners[2] = CPoint(d, -1.2, +1.2);
        corners[3] = CPoint(d, -1.2, -1.2);
    }
    else if (n[1] == 1)
    {
        corners[0] = CPoint(-1.2, d, +1.2);
        corners[1] = CPoint(+1.2, d, +1.2);
        corners[2] = CPoint(+1.2, d, -1.2);
        corners[3] = CPoint(-1.2, d, -1.2);
    }
    else if (n[2] == 1)
    {
        corners[0] = CPoint(+1.2, -1.2, d);
        corners[1] = CPoint(+1.2, +1.2, d);
        corners[2] = CPoint(-1.2, +1.2, d);
        corners[3] = CPoint(-1.2, -1.2, d);
    }

    glBegin(GL_POLYGON);
    glNormal3d(n[0], n[1], n[2]);
    for (int i = 0; i < 4; ++i)
    {
        glVertex3d(corners[i][0], corners[i][1], corners[i][2]);
    }
    glEnd();

    glEnable(GL_CULL_FACE);
    glDisable(GL_BLEND);
}

/*! draw mesh */
void draw_mesh()
{

    if (g_show_sites)
       draw_sites();
    
    if (g_show_voronoi_vertices)
        draw_voronoi_vertices();
    
    draw_voronoi_cells();

    switch (displayMode)
    {
        case 0:
            draw_halffaces(g_tmesh.halffaces_below(), GL_FILL);
            break;
        case 1:
            draw_halffaces(g_tmesh.halffaces_below(), GL_LINE);
            break;
        case 2:
            draw_halffaces(g_tmesh.halffaces_below(), GL_FILL);
            draw_halffaces(g_tmesh.halffaces_above(), GL_LINE);
            break;
    }

    draw_plane();
}

/*! display call back function
 */
void display()
{
    /* clear frame buffer */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    setupLight();
    /* transform from the eye coordinate system to the world system */
    setupEye();
    glPushMatrix();
    /* transform from the world to the ojbect coordinate system */
    setupObject();
    /* draw the mesh */
    draw_mesh();

    glPopMatrix();
    glutSwapBuffers();
}

/*! Called when a "resize" event is received by the window. */

void reshape(int w, int h)
{
    float ar;

    win_width = w;
    win_height = h;

    ar = (float) (w) / h;
    glViewport(0, 0, w, h); /* Set Viewport */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    // magic imageing commands
    gluPerspective(40.0, /* field of view in degrees */
                   ar,   /* aspect ratio */
                   1.0,  /* Z near */
                   100.0 /* Z far */);

    glMatrixMode(GL_MODELVIEW);

    glutPostRedisplay();
}

/*! helper function to remind the user about commands, hot keys */
void help()
{
    printf("V  -  compute the voronoi diagram for the input sites\n");
    printf("i  -  increase the selected dual cell id(vertex id) to show\n");
    printf("d  -  decrease the selected dual cell id(vertex id) to show\n");
    printf("g  -  switch display mode for delaunay triangulation\n");
    printf("s  -  show or hidden the sites\n");
    printf("v  -  show or hidden the voronoi vertices\n");
    printf("+  -  shift the cutting plane upwards\n");
    printf("-  -  shift the cutting plane downwards\n");
    printf("x  -  set the cutting plane normal (1,0,0)\n");
    printf("y  -  set the cutting plane normal (0,1,0)\n");
    printf("z  -  set the cutting plane normal (0,0,1)\n");
    printf("?  -  help Information\n");
    printf("esc - quit\n");
}

/*! Keyboard call back function */

void keyBoard(unsigned char key, int x, int y)
{
    clock_t begin = clock(), end;
    switch (key)
    {
        case 'i':
            g_selected_vertex_id++;
            printf("cell id: %d\n", g_selected_vertex_id);
            break;
        case 'd':
            g_selected_vertex_id--;
            printf("cell id: %d\n", g_selected_vertex_id);
            break;
        case 'V':
            // compute the delaunay triangulation
            begin = clock();
            g_pd.calc_delaunay();
            end = clock();
            printf("Computing DT takes %g sec.\n", (double) (end - begin) / CLOCKS_PER_SEC);
            g_tmesh.cut(CPlane(n, d));

            // compute the power diagram
            begin = clock();
            g_pd.calc_voronoi();
            end = clock();
            printf("Computing PD takes %g sec.\n", (double) (end - begin) / CLOCKS_PER_SEC);

            break;
        case 's':
            g_show_sites = !g_show_sites;
            break;
        case 'v':
            g_show_voronoi_vertices = !g_show_voronoi_vertices;
            break;
        case '+':
            d += 0.05;
            g_tmesh.cut(CPlane(n, d));
            break;
        case '-':
            d -= 0.05;
            g_tmesh.cut(CPlane(n, d));
            break;
        case 'x':
            n = CPoint(1, 0, 0);
            g_tmesh.cut(CPlane(n, d));
            break;
        case 'y':
            n = CPoint(0, 1, 0);
            g_tmesh.cut(CPlane(n, d));
            break;
        case 'z':
            n = CPoint(0, 0, 1);
            g_tmesh.cut(CPlane(n, d));
            break;
        case 'g':
            displayMode = (displayMode + 1) % 3;
            break;
        case 'r':
            switchDirection = !switchDirection;
            if (switchDirection)
                glFrontFace(GL_CW);
            else
                glFrontFace(GL_CCW);
            break;
        case '?':
            help();
            break;
        case 27:
            exit(0);
            break;
    }

    glutPostRedisplay();
}

/*! setup GL states */
void setupGLstate()
{
    GLfloat lightOneColor[] = {1.f, 1.f, 1.f, 1.f};
    GLfloat globalAmb[] = {.1f, .1f, .1f, 1.f};
    GLfloat lightOnePosition[] = {.0f, .0f, 1.f, 0.f};

    glEnable(GL_CULL_FACE);
    if (switchDirection)
        glFrontFace(GL_CW);
    else
        glFrontFace(GL_CCW);

    glEnable(GL_DEPTH_TEST);
    glClearColor(0.3f, 0.3f, 0.8f, 0.3f);
    glShadeModel(GL_SMOOTH);

    glEnable(GL_LIGHT1);
    glEnable(GL_LIGHTING);
    glEnable(GL_NORMALIZE);
    glEnable(GL_COLOR_MATERIAL);

    glLightfv(GL_LIGHT1, GL_DIFFUSE, lightOneColor);
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, globalAmb);

    glLightfv(GL_LIGHT1, GL_POSITION, lightOnePosition);

    GLfloat diffuseMaterial[4] = {0.5, 0.5, 0.5, 1.0};
    GLfloat mat_specular[] = {1.0, 1.0, 1.0, 1.0};

    glEnable(GL_DEPTH_TEST);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, diffuseMaterial);
    glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
    glMaterialf(GL_FRONT, GL_SHININESS, 25.0);
    glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
}

/*! mouse click call back function */
void mouseClick(int button, int state, int x, int y)
{
    /* set up an arcball around the Eye's center
            switch y coordinates to right handed system  */

    if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
    {
        gButton = GLUT_LEFT_BUTTON;
        arcball = CArcball(win_width, win_height, x - win_width / 2, win_height - y - win_height / 2);
    }

    if (button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN)
    {
        startx = x;
        starty = y;
        gButton = GLUT_MIDDLE_BUTTON;
    }

    if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN)
    {
        startx = x;
        starty = y;
        gButton = GLUT_RIGHT_BUTTON;
    }
    return;
}

/*! mouse motion call back function */
void mouseMove(int x, int y)
{
    CPoint trans;
    CQrot rot;

    /* rotation, call arcball */
    if (gButton == GLUT_LEFT_BUTTON)
    {
        rot = arcball.update(x - win_width / 2, win_height - y - win_height / 2);
        ObjRot = rot * ObjRot;
        glutPostRedisplay();
    }

    /*xy translation */
    if (gButton == GLUT_MIDDLE_BUTTON)
    {
        double scale = 10. / win_height;
        trans = CPoint(scale * (x - startx), scale * (starty - y), 0);
        startx = x;
        starty = y;
        ObjTrans = ObjTrans + trans;
        glutPostRedisplay();
    }

    /* zoom in and out */
    if (gButton == GLUT_RIGHT_BUTTON)
    {
        double scale = 10. / win_height;
        trans = CPoint(0, 0, scale * (starty - y));
        startx = x;
        starty = y;
        ObjTrans = ObjTrans + trans;
        glutPostRedisplay();
    }
}

int main(int argc, char* argv[])
{
    //int nSites = 10000;
    int nSites = 100;
    if (argc > 1)
        nSites = atoi(argv[1]);
    
    g_pd.init(nSites);

    /* glut stuff */
    glutInit(&argc, argv); /* Initialize GLUT */
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
    glutInitWindowSize(600, 600);
    glutCreateWindow("Tet Mesh Viewer"); /* Create window with given title */
    glViewport(0, 0, 800, 800);

    glutDisplayFunc(display); /* Set-up callback functions */
    glutReshapeFunc(reshape);
    glutMouseFunc(mouseClick);
    glutMotionFunc(mouseMove);
    glutKeyboardFunc(keyBoard);
    setupGLstate();

    glutMainLoop(); /* Start GLUT event-processing loop */

    return 0;
}
